from astropy.coordinates import SkyCoord, SphericalCosLatDifferential, Distance
from astropy.coordinates import get_sun, SphericalRepresentation, SkyOffsetFrame, ICRS
from astropy.time import Time
from astropy.table import Table
import astropy.units as u
from astroquery.vizier import Vizier
import warnings
import numpy as np
from .config import test_attr

### Object for star
def search_star(**kwargs):
    """ Search position on Vizier and returns a catalogue.

    Parameters:
        coord (str, astropy.SkyCoord):Coordinate to search around.
        code (str):Gaia Source_id of the star
        columns (list):List of strings with the name of the columns to retrieve.
        radius (int, float, unit.quantity):Radius to search around coord.
        catalog (str):Vizier catalogue to search.

    Returns:
        catalogue(astropy.Table):An astropy Table with the catalogue informations.   

    """
    row_limit = 100
    if 'log' in kwargs and kwargs['log']:
        print('Downloading star parameters from {}'.format(kwargs['catalog']))
    vquery = Vizier(columns=kwargs['columns'], row_limit=row_limit, timeout=600)
    if 'code' in kwargs:
        catalogue = vquery.query_constraints(catalog=kwargs['catalog'], Source=kwargs['code'])
    elif 'coord' in kwargs:
        catalogue = vquery.query_region(kwargs['coord'], radius=kwargs['radius'], catalog=kwargs['catalog'])
    else:
        raise ValueError('At least a code or coord should be given as input')
    return catalogue


def van_belle(magB=None, magV=None, magK=None):
    '''
    Determine the diameter of a star in mas using equations from van Belle (1999) 
    -- Publi. Astron. Soc. Pacific 111, 1515-1523:

    Inputs:
        magB, magV, magK: The magnitudes B, V and K of the star
    '''
    if magB is None or np.isnan(magB) or magB > 49:
        magB = np.nan
    if magV is None or np.isnan(magV) or magV > 49:
        magV = np.nan
    if magK is None or np.isnan(magK) or magK > 49:
        magK = np.nan

    def calc_diameter(a1, a2, mag):
        return 10**(a1 + a2*(mag - magK) - 0.2*mag)

    params = {'sg': {'B': [ 0.648, 0.220], 'V': [0.669, 0.223]},
              'ms': {'B': [ 0.500, 0.290], 'V': [0.500, 0.264]},
              'vs': {'B': [ 0.840, 0.211], 'V': [0.789, 0.218]}}

    mag = np.array([magB, magV])
    diameter = {}
    for st in ['sg', 'ms', 'vs']:
        diameter_s = {}
        for i,m in enumerate(['B','V']):
            diam = calc_diameter(*params[st][m], mag[i])
            if not np.isnan(diam):
                diameter_s[m] = calc_diameter(*params[st][m], mag[i])*u.mas
        if diameter_s:
            diameter[st] = diameter_s
    return diameter
    
    
def kervella(magB=None, magV=None, magK=None):
    '''
    Determine the diameter of a star in mas using equations from Kervella et. al (2004) 
    -- A&A Vol.  426, No.  1:

    Inputs:
        magB, magV, magK: The magnitudes B, V and K of the star
    '''
    if magB is None or np.isnan(magB) or magB > 49:
        magB = np.nan
    if magV is None or np.isnan(magV) or magV > 49:
        magV = np.nan
    if magK is None or np.isnan(magK) or magK > 49:
        magK = np.nan
    const1 = np.array([0.0755, 0.0535])
    const2 = np.array([0.5170, 0.5159])
    mag = np.array([magV,magB])
    vals = 10**(const1*(mag-magK)+const2-0.2*magK)
    diam = {}
    if not np.isnan(vals[0]):
        diam['V'] = vals[0]*u.mas
    if not np.isnan(vals[1]):
        diam['B'] = vals[1]*u.mas
    return diam


class Star():
    def __init__(self,**kwargs):
        '''Define a star

        Parameters:
            code (str): Gaia-DR2 Source code for searching in Vizier.
            coord (str, SkyCoord): if code is not given, coord nust have
                the coordinate to search in Vizier: 'hh mm ss.ss +dd mm ss.ss'
            nomad (bool): If true, it tries to download the magnitudes
                from NOMAD catalogue.
            log (bool): If true, it prints the downloaded information.
            local (bool): If true, it uses the given coordinate in 'coord'
                as final coordinate.
        '''
        self.__local = False
        self.mag = {}
        self.errors = {}
        self.__log = True
        if 'log' in kwargs:
            self.__log = kwargs['log']
        if 'local' in kwargs:
            self.__local = test_attr(kwargs['local'], bool, 'local')
        if not any(i in kwargs for i in ['coord', 'code']):
            raise KeyError('Input values must have either the coordinate of the star or the Gaia code for online search')
        if 'coord' in kwargs:
            self.coord = SkyCoord(kwargs['coord'], unit=('hourangle', 'deg'))
            #self.coord = test_attr(kwargs['coord'], SkyCoord, 'coord')
        else:
            self.__local=False
        if 'code' in kwargs:
            self.code = test_attr(kwargs['code'], str, 'code')
        if not self.__local:
            self.__searchgaia()
        nomad = kwargs.get('nomad', True)
        if nomad:
            self.__getcolors()
    
    
    def set_magnitude(self,**kwargs):
        '''
        Set the magnitudes of a star.

        Inputs:
            (band name): The magnitude for given band

        Example:
            set_magnitude(G=10)
            set_magnitude(K=15)
            set_magnitude(newband=6)
        '''
        for key in kwargs:
            mag = test_attr(kwargs[key], float, key)
            if key in self.mag:
                warnings.warn('{0} mag already defined. {0}={1} will be replaced by {0}={2}'.format(key, self.mag[key], kwargs[key]))
            self.mag[key] = kwargs[key]

            
    def set_diameter(self, diameter):
        '''
        Set a user diameter of a star in mas.

        Inputs:
            diameter = float, in mas
        '''
        self.diameter_user = diameter*u.mas

    
    def van_belle(self):
        '''
        Determine the diameter of a star in mas using equations from van Belle (1999) 
        -- Publi. Astron. Soc. Pacific 111, 1515-1523:
       '''
        return van_belle(self.mag.get('B'), self.mag.get('V'), self.mag.get('K'))
    

    def kervella(self):
        '''
        Determine the diameter of a star in mas using equations from Kervella et. al (2004) 
        -- A&A Vol.  426, No.  1:
        '''
        return kervella(self.mag.get('B'), self.mag.get('V'), self.mag.get('K'))
    
    
    def apparent_diameter(self, distance, mode='auto', log=True, **kwargs):
        """Calculate the apparent diameter of the star at given distance

        Parameters:
            distance (int, float): Object distance in AU
            mode: The way to calculate the apparent diameter
                'user': calculates using user diameter
                'gaia': calculates using Gaia diameter
                'kervella': calculates using Kervella equations
                'van_belle': calculates using van Belle equations
                'auto' (default): tries all the above methods
                    until it is able to calculate diameter.
                    The order of try is the same as shows above.
            'band': If mode is 'kervella' or 'van_belle',
                the filter must be given, 'B' or 'V'.
                In 'auto' mode, 'V' is selected.
            'star_type': If mode is 'van_belle', the star type must be given.
                It can be 'sg', 'ms' and 'vs' for 'Super Giant', 'Main
                Sequence' and 'Variable Star'. In 'auto' mode, 'sg' is selected.
            'log': If True, it prints the mode used by 'auto'.
        """
        try:
            d = distance.to(u.km)
        except:
            distance = distance*u.AU

        kwargs['band'] = kwargs.get('band', 'V')
        kwargs['star_type'] = kwargs.get('star_type', 'sg')

        if mode in ['user', 'auto']:
            try:
                diam = distance*np.tan(self.diameter_user)
                if log:
                    print('Calculating apparent diameter from user defined diameter')
                return diam.to(u.km)
            except:
                pass
        
        if mode == 'user':
            raise ValueError('User diameter must be informed.')
            
        if mode in ['gaia', 'auto']:
            try:
                diam = distance*np.tan(self.diameter_gaia)
                if log:
                    print('Apparent diameter using Gaia')
                return diam.to(u.km)
            except:
                pass
                
        if mode == 'gaia':
            raise ValueError('It is not possible to calculate star diameter from Gaia.')
                
        if kwargs['band'] not in ['B', 'V']:
            raise KeyError('band must be informed as "B", or "V"')
            
        if mode in ['kervella', 'auto']:
            diam_kerv = self.kervella().get(kwargs['band'])
            if diam_kerv is None:
                raise ValueError('Diameter could not be calculated for given band')
            if log:
                print('Apparent diameter using Kervella et al. (2004)')
            diam = distance*np.tan(diam_kerv)
            return diam.to(u.km)
            
        if kwargs['star_type'] not in ['sg', 'ms', 'vs']:
            raise KeyError('star_type must be informed as "sg", "ms" or "vs"')
            
        if mode in ['van_belle', 'auto']:
            diam_van = self.van_belle().get(kwargs['star_type'])
            if diam_van is None:
                raise ValueError('Diameter could not be calculated using Van Belle')
            diam_van = diam_van.get(kwargs['band'])
            if diam_van is None:
                raise ValueError('Diameter could not be calculated for given band')
            if log:
                print('Apparent diameter using van Belle (1999)')
            diam = distance*np.tan(diam_van)
            return diam.to(u.km)
                        
        raise AttributeError("Star apparent diameter could not be calculated. ",
                             "Please define star diameter or B,V,K magnitudes.")
        
    
    def __searchgaia(self):
        """search for the star position in the gaia catalogue and save informations
        """
        columns = ['Source', 'RA_ICRS', 'e_RA_ICRS', 'DE_ICRS', 'e_DE_ICRS', 'Plx', 'pmRA', 'e_pmRA', 'pmDE', 'e_pmDE', 'Gmag', 'e_Gmag', 'Dup', 'Epoch', 'Rad']
        if hasattr(self, 'code'):
            catalogue = search_star(code=self.code, columns=columns, catalog='I/345/gaia2', log=self.__log)
        else:
            catalogue = search_star(coord=self.coord, columns=columns, radius=2*u.arcsec, catalog='I/345/gaia2', log=self.__log)
        if len(catalogue) == 0:
            raise ValueError('No star was found within 2 arcsec from given coordinate')
        catalogue = catalogue[0]
        if len(catalogue) > 1:
            if self.__log:
                print('{} stars were found within 2 arcsec from given coordinate.'.format(len(catalogue)))
                print('The list below is sorted by distance. Please select the correct star')
            gmag = catalogue['Gmag']
            tstars = SkyCoord(catalogue['RA_ICRS'], catalogue['DE_ICRS'])
            sep = tstars.separation(self.coord)
            k = sep.argsort()
            while True:
                t = Table()
                t['num'] = np.arange(len(tstars))+1
                t['dist(")'] = sep[k].arcsec
                t['dist(")'].format = '6.3f'
                t['Gmag'] = gmag[k].quantity.value
                t['Gmag'].format = '6.3f'
                t['RA___ICRS___DEC'] = tstars[k].to_string('hmsdms', precision=4)
                t.pprint_all()
                print('  0: Cancel')
                choice = int(input('Choose the corresponding number of the correct star: '))
                if choice in np.arange(len(k)+1):
                    break
                print('{} is not a valid choice. Please select the correct star'.format(choice))
            if choice == 0:
                raise ValueError('It was not possible to define a star')
            catalogue = catalogue[[k[choice-1]]]
        self.code = catalogue['Source'][0]
        ra = catalogue['RA_ICRS']
        dec = catalogue['DE_ICRS']
        pmra = catalogue['pmRA']
        pmde = catalogue['pmDE']
        epoch = Time(catalogue['Epoch'].quantity, format='jyear')
        try:
            distance = Distance(parallax=catalogue['Plx'].quantity, allow_negative=False)
            self.coord = SkyCoord(ra, dec, distance=distance, pm_ra_cosdec=pmra, pm_dec=pmde, obstime=epoch[0])[0]
        except:
            self.coord = SkyCoord(ra, dec, pm_ra_cosdec=pmra, pm_dec=pmde, obstime=epoch[0])[0]
        self.set_magnitude(G=catalogue['Gmag'][0])
        self.errors['RA'] = catalogue['e_RA_ICRS'][0]*u.mas
        self.errors['DEC'] = catalogue['e_DE_ICRS'][0]*u.mas
        self.errors['pmRA'] = catalogue['e_pmRA'][0]*(u.mas/u.yr)
        self.errors['pmDEC'] = catalogue['e_pmDE'][0]*(u.mas/u.yr)
        rad = catalogue['Rad'][0]
        if np.ma.core.is_masked(rad) or np.ma.core.is_masked(catalogue['Plx'][0]):
            if self.__log:
                warnings.warn('Gaia catalogue does not have star radius.')
        else:
            self.diameter_gaia = 2*np.arctan((rad*u.solRad)/distance[0]).to(u.mas)
        if self.__log:
            print('1 Gaia-DR2 star found G={}'.format(catalogue['Gmag'][0]))
            print('star coordinate at J{}: RA={} +/- {}, DEC={} +/- {}'.format(self.coord.obstime.jyear,
                  self.coord.ra.to_string(u.hourangle, sep='hms', precision=5), self.errors['RA'],
                  self.coord.dec.to_string(u.deg, sep='dms', precision=4), self.errors['DEC']))
        
        
    def __getcolors(self):
        """search for the B,V,K magnitudes of the star on Vizier and saves the result
        """
        columns = ['RAJ2000', 'DEJ2000', 'Bmag', 'Vmag', 'Rmag', 'Jmag', 'Hmag', 'Kmag']
        catalogue = search_star(coord=self.coord, columns=columns, radius=2*u.arcsec, catalog='I/297/out', log=self.__log)
        if len(catalogue) == 0:
            if self.__log:
                warnings.warn('No star was found on NOMAD that matches the star')
            return
        catalogue = catalogue[0]
        if len(catalogue) > 1:
            print('{} stars were found within 2 arcsec from given coordinate.'.format(len(catalogue)))
            print('The list below is sorted by distance. Please select the correct star')
            print('Star G mag: {}'.format(self.mag['G']))
            tstars = SkyCoord(catalogue['RAJ2000'], catalogue['DEJ2000'])
            sep = tstars.separation(self.coord)
            k = sep.argsort()
            while True:
                t = Table()
                t['num'] = np.arange(len(tstars))+1
                t['dist(")'] = sep[k].arcsec
                t['dist(")'].format = '6.3f'
                t['Bmag'] = catalogue['Bmag'][k].quantity.value
                t['Vmag'] = catalogue['Vmag'][k].quantity.value
                t['Rmag'] = catalogue['Rmag'][k].quantity.value
                t['Jmag'] = catalogue['Jmag'][k].quantity.value
                t['Hmag'] = catalogue['Hmag'][k].quantity.value
                t['Kmag'] = catalogue['Kmag'][k].quantity.value
                t['Bmag'].format = t['Vmag'].format = t['Rmag'].format = t['Jmag'].format = t['Hmag'].format = t['Kmag'].format = '6.3f'
                t['RA___ICRS___DEC'] = tstars[k].to_string('hmsdms', precision=4)
                t.pprint_all()
                print('  0: Cancel')
                choice = int(input('Choose the corresponding number of the correct star: '))
                if choice in np.arange(len(k)+1):
                    break
                print('{} is not a valid choice. Please select the correct star'.format(choice))
            if choice == 0:
                warnings.warn('No magnitudes were obtained from NOMAD')
                return
            catalogue = catalogue[[k[choice-1]]]
        errors = []
        for mag in ['B', 'V', 'R', 'J', 'H', 'K']:
            name = mag + 'mag'
            if np.ma.core.is_masked(catalogue[name][0]):
                errors.append(mag)
                continue
            self.set_magnitude(**{mag: catalogue[name][0]})
        if len(errors) > 0 and self.__log:
            warnings.warn('Magnitudes in {} were not located in NOMAD'.format(errors))
        
        
    def geocentric(self, time):
        """ calculate the position of the star using parallax and proper motion

        Parameters:
            time (float, Time):time to apply proper motion and calculate paralax.
        """
        try:
            time = Time(time)
        except:
            time = Time(time, format='jd', scale='utc')
        n_coord = self.barycentric(time)
        if self.coord.distance.unit.is_unity() or np.isnan(self.coord.distance):
            g_coord = n_coord
        else:
            sun = get_sun(time)
            g_coord = SkyCoord(*(n_coord.cartesian.xyz + sun.cartesian.xyz), representation_type='cartesian')
            g_coord = g_coord.represent_as(SphericalRepresentation)
            g_coord = SkyCoord(g_coord.lon, g_coord.lat, g_coord.distance)

        if hasattr(self, 'offset'):
            star_frame = SkyOffsetFrame(origin=g_coord)
            new_pos = SkyCoord(lon=self.offset.d_lon_coslat, lat=self.offset.d_lat, frame=star_frame)
            return new_pos.transform_to(ICRS)
        
        return g_coord
    
    
    def barycentric(self, time):
        """ calculate the position of the star using proper motion
        
        Parameters:
            time (float, Time):time to apply proper motion.
        """
        try:
            time = Time(time)
        except:
            time = Time(time, format='jd', scale='utc')
        if np.isnan(self.coord.pm_dec):
            return self.coord
        if self.coord.distance.unit.is_unity():
            star_frame = SkyOffsetFrame(origin=self.coord)
            dt = time - self.coord.obstime
            new_pos = SkyCoord(lon=self.coord.pm_ra_cosdec*dt, lat=self.coord.pm_dec*dt, frame=star_frame)
            n_coord = new_pos.transform_to(ICRS)
        else:
            n_coord = self.coord.apply_space_motion(new_obstime=time)
        return n_coord

    def error_at(self, time):
        """Estimates star position error at time given

        Parameters:
            time (float, Time):time to apply proper motion.

        Return:
            pair of errors, in RA* and DEC
        """
        try:
            time = Time(time)
        except:
            time = Time(time, format='jd', scale='utc')
        e_ra = self.errors['RA'] + self.errors['pmRA']*np.abs(time-self.coord.obstime)
        e_dec = self.errors['DEC'] + self.errors['pmDEC']*np.abs(time-self.coord.obstime)
        return e_ra, e_dec

    def add_offset(self, da_cosdec, ddec):
        """Add an offset to star

        Parameters:
            da_cosdec (int, float):Delta_alpha_cos_delta in mas
            ddec (int, float):Delta_delta in mas
        """
        dadc = test_attr(da_cosdec, float, 'da_cosdec')
        dd = test_attr(ddec, float, 'ddec')
        self.offset = SphericalCosLatDifferential(dadc*u.mas, dd*u.mas, 0.0*u.km)

    def __str__(self):
        """String representation of the Star class
        """
        out = 'ICRS star coordinate at J{}:\nRA={} +/- {:.4f}, DEC={} +/- {:.4f}\n\n'.format(self.coord.obstime.jyear,
            self.coord.ra.to_string(u.hourangle, sep='hms', precision=5), self.errors['RA'],
            self.coord.dec.to_string(u.deg, sep='dms', precision=4), self.errors['DEC'])
        if hasattr(self, 'code'):
            out += 'Gaia-DR2 star Source ID: {}\n'.format(self.code)
        if hasattr(self, 'offset'):
            out += 'Offset Apllied: d_alpha_cos_dec = {}, d_dec = {}\n'.format(self.offset.d_lon_coslat, self.offset.d_lat)
        out += 'Magnitudes:'
        mag_out = [' {}: {:6.3f}'.format(mag, self.mag[mag]) for mag in self.mag]
        out_mag = []
        for i,mag in enumerate(mag_out):
            if i%6 == 0:
                out_mag.append([])
            out_mag[-1].append(mag)
        out += (',\n'+' '*11).join([','.join(out_i) for out_i in out_mag])
        out += '\n\n'
        if hasattr(self, 'diameter_gaia'):
            out += 'Apparent diameter: {:.4f}, Source: Gaia-DR2\n'.format(self.diameter_gaia)
        if hasattr(self, 'diameter_user'):
            out += 'Apparent diameter: {:.4f}, Source: User\n'.format(self.diameter_user)
        kerv = self.kervella()
        if kerv:
            out += 'Apparent diameter from Kervella et. al (2004):\n'
            out += '   ' + ','.join([' {}: {:.4f}'.format(k, v) for k,v in kerv.items()])
        vanb = self.van_belle()
        if vanb:
            out += '\nApparent diameter from van Belle (1999):'
            for key, value in vanb.items():
                out += '\n    {}:'.format(key)
                out += ','.join([' {}: {:.4f}'.format(k, v) for k,v in value.items()])
        return out



from astropy.coordinates import SkyCoord, EarthLocation, GCRS
from astropy.coordinates.matrix_utilities import rotation_matrix
from astropy.time import Time
import astropy.units as u
from astroquery.mpc import MPC
import numpy as np
from .config import test_attr
from .star import Star

def search_code_mpc():
    """ Reads the MPC Observer Database

    Return:
        observatories (dict): A python dictionaty with all the sites
            as an Astropy EarthLocation object
    """
    obs = MPC.get_observatory_codes()
    observatories = {}
    for line in obs:
        code = line['Code']
        lon = line['Longitude']*u.deg
        rcphi = line['cos']*6378.137*u.km
        rsphi = line['sin']*6378.137*u.km
        name = line['Name']
        site = EarthLocation.from_geocentric(rcphi*np.cos(lon), rcphi*np.sin(lon), rsphi)
        observatories[code] = (name, site)
    return observatories
        

class Observer():
    """Define the observer object
    """
    __names = []
    def __init__(self, *args, **kwargs):
        """
        Parameters:
            name (str): Name for the Observer. (required)
                Each time an Observer object is defined, the name must be different.
            code (str): The IAU code to look for coordinates in MPC database
            site (EarthLocation): To provide a EarthLocation object to Object.
            lon (int, float): The Longitude of the site.
            lat (int, float): The Latitude of the site.
            height (int, float): The height of the site.

        The user must provide one of the followings:
        Observer(code)
        Observer(name, code) if the user wants to use a different name from the MPC database
        Observer(name, site)
        Observer(name, lon, lat, height)
        """
        if len(args) == 1 or 'code' in kwargs:
            if len(args) == 1:
                code = args[0]
                try:
                    code = test_attr(args[0], float, 'code')
                    code = '{:03.0f}'.format(code)
                    kwargs['code'] = code
                except:
                    pass
            self.code = kwargs['code']
            try:
                name, self.site = search_code_mpc()[self.code]
                self.__name = kwargs.get('name', name)
            except:
                raise ValueError('code {} could not be located in MPC database'.format(self.code))
        elif len(args) == 2 or all(i in kwargs for i in ['name', 'site']):
            if len(args) == 2:
                kwargs['name'] = args[0]
                kwargs['site'] = args[1]
            self.__name = kwargs['name']
            self.site = test_attr(kwargs['site'], EarthLocation, 'site')
        elif len(args) == 4 or all(i in kwargs for i in ['name', 'lon', 'lat', 'height']):
            if len(args) == 4:
                kwargs['name'] = args[0]
                kwargs['lon'] = args[1]
                kwargs['lat'] = args[2]
                kwargs['height'] = args[3]
            self.__name = kwargs['name']
            self.site = EarthLocation(kwargs['lon'], kwargs['lat'], kwargs['height'])
        else:
            raise ValueError('Input parameters could not be determined')
        if self.__name in self.__names:
            raise ValueError('name {} already defined for another Observer object. Please choose a different one.'.format(self.__name))
        self.__names.append(self.__name)
        
    def get_ksi_eta(self, time, star):
        """ Calculates relative position to star in the orthographic projection.
        
        Parameters:
        time (str, Time):Time from which to calculate the position.
        It can be a string in the format "yyyy-mm-dd hh:mm:ss.s" or an astropy Time object
        star (str, SkyCoord):The coordinate of the star in the same frame as the ephemeris.
        It can be a string in the format "hh mm ss.s +dd mm ss.ss" or an astropy SkyCoord object.
        
        Returns:
        ksi, eta (float): on-sky orthographic projection of the observer relative to a star
        """
        time = test_attr(time, Time, 'time')
        try:
            star = SkyCoord(star, unit=(u.hourangle,u.deg))
        except:
            raise ValueError('star is not an astropy object or a string in the format "hh mm ss.s +dd mm ss.ss"')

        itrs = self.site.get_itrs(obstime=time)
        gcrs = itrs.transform_to(GCRS(obstime=time))
        rz = rotation_matrix(star.ra, 'z')
        ry = rotation_matrix(-star.dec, 'y')

        cp = gcrs.cartesian.transform(rz).transform(ry)
        return cp.y.to(u.km).value, cp.z.to(u.km).value
    
    def sidereal_time(self, time, mode='local'):
        """Calculates the Apparent Sidereal Time at a certain time

        Parameters:
            time (str,Time): Time to calculate sidereal time.
            mode (str): if 'local' it calculates the sidereal time for
                the coordinates of this object. If 'greenwich', it
                calculates the Greenwich Apparent Sidereal Time.

        Returns:
            sidereal_time: An Astropy Longitude object with the ST.
        """
        # return local or greenwich sidereal time
        time = test_attr(time, Time, 'time')
        time.location = self.site
        if mode == 'local':
            return time.sidereal_time('apparent')
        elif mode == 'greenwich':
            return time.sidereal_time('apparent', 'greenwich')
        else:
            raise ValueError('mode must be "local" or "greenwich"')

    @property
    def name(self):
        return self.__name

    @property
    def lon(self):
        return self.site.lon

    @lon.setter
    def lon(self, lon):
        lat = self.site.lat
        height = self.site.height
        site = EarthLocation(lon, lat, height)
        self.site = site

    @property
    def lat(self):
        return self.site.lat

    @lat.setter
    def lat(self, lat):
        lon = self.site.lon
        height = self.site.height
        site = EarthLocation(lon, lat, height)
        self.site = site

    @property
    def height(self):
        return self.site.height

    @height.setter
    def height(self, height):
        lon = self.site.lon
        lat = self.site.lat
        site = EarthLocation(lon, lat, height)
        self.site = site

    def __str__(self):
        """String representation of the Observer class
        """
        out = 'Site: {}\n'.format(self.name)
        out += 'Geodetic coordinates: Lon: {}, Lat: {}, height: {:.3f}'.format(
            self.site.lon.__str__(), self.site.lat.__str__(), self.site.height.to(u.km))
        return out

    def __del__(self):
        """ When this object is deleted, it removes the name from the Class name list.
        """
        try:
            self.__names.remove(self.__name)
        except:
            pass



### Useful classes
class colors():
    '''
    Docstring
    '''
    __cores = {
    'positive': 'blue',
    'negative': 'green',
    'error': 'red'
    }
    def __init__(self):
        pass

    @property
    def positive_color(self):
        return self.__cores['positive']

    @positive_color.setter
    def positive_color(self, value):
        self.__cores['positive']=value

    @property
    def negative_color(self):
        return self.__cores['negative']

    @negative_color.setter
    def negative_color(self, value):
        self.__cores['negative']=value

    @property
    def error_bar(self):
        return self.__cores['error']

    @error_bar.setter
    def error_bar(self, value):
        self.__cores['error']=value


### useful functions
def test_attr(attr, typ, name):
    """
    This function tests if the attribute "attr" belongs to the type "typ",
    if not it gives an error message informing the name of the variable
    """
    try:
        return typ(attr)
    except:
        raise TypeError('"{}" keyword must be a {} object'.format(name, typ))


praia_occ_head = (" Notes:\n"
                " C/A: geocentric closest approach, in arcsec\n"
                " P/A: Planet position angle wrt to star at C/A, in deg\n"
                " vel: velocity in plane of sky, in km/sec, positive= prograde, negative= retrograde\n"
                " R: instrumental magnitude in UCAC2 system\n"
                " J, H, K: 2MASS magnitudes (50.0 = not in 2MASS)\n"
                " R*, J*, H*, K* are normalized magnitudes to a common\n"
                " shadown velocity of 20 km/sec by the relationship:\n"
                " Mag* = Mag_actual + 2.5*log10[velocity/20 (km/sec)]\n"
                " Delta: Planet range to Earth, AU\n"
                " long: East longitude of sub-planet point, deg, positive towards East\n"
                " loc. t.= UT + long: local solar time at sub-planet point, hh:mm\n"
                " --------------------------------------------------------------------------------------------------------------------\n"
                " Selection criteria:\n"
                " Maximum geocentrique closest approach considered:   {max_ca:.3f}\n"
                " Day light exclusion (range loc. t.): 12.0hs to 12.0 hs (t1 = t2 -> no exclusions)\n"
                " List below has:           {size}  entries\n"
                " Reference ephemeris: {ephem}\n"
                " Offset applied to ephemeris off_ra(mas) = A * (t-t0) + B \n"
                " Offset applied to ephemeris off_de(mas) = C * (t-t0) + D \n"
                " t0 =    2005.0000000000000       yrs\n"
                " A =    0.0000000000000000       (mas/yr)\n"
                " B =    0.0000000000000000       (mas)\n"
                " C =    0.0000000000000000       (mas/yr)\n"
                " D =    0.0000000000000000       (mas)\n"
                " pm = proper motion applied? (ok, no)\n"
                " ct = uc (UCAC2); 2m (2MASS); fs (field star); g2 (Gaia-DR2)\n"
                " f = multiplicity flag\n"
                " 0 - no multiple entries per star in astrometry\n"
                " 1 - single position from 2 or more uc/2m entries\n"
                " 2 - single position from 1 uc/2m entry only\n"
                " 3 - fs position from entry with more N contributions\n"
                " 4 - fs position from entry with best (x,y) error\n"
                " 5 - fs position from entry with brightest R mag.\n"
                " 6 - fs position from average over all entries\n"
                " (see details in Assafin et al. 2009)\n"
                " E_ra, E_de: error of star position (mas); (9999 = no estimation)\n"
                " pmra, pmde: star proper motion (mas); (9999 = no proper motion)\n"
                " ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n"
                "  d  m year  h:m:s UT     ra___dec___J2000_candidate     ra_dec_J2000_target_geocen     C/A    P/A     vel  Delta  G*   J*   H*   K*   long  loc. t.  off_ra   off_de pm ct f E_ra E_de pmra pmde\n"
                " ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------"
                 )


ow_occ_head = (" Planete: {name}: Star GAIA-DR2, {ephem}\n"
            " Notes:\n"
            " C/A: geocentric closest approach, in arcsec\n"
            " P/A: Planet's position angle wrt to star at C/A, in deg\n"
            " vel: velocity in plane of sky, in km/sec, positive= prograde, negative= retrograde\n"
            " G*, K*: quite rough R and K magnitudes from 1.2m ESO Swiss telescope,\n"
            "         **corrected to a standard velocity of 20 km/sec**\n"
            " Delta: Planet's range to Earth, AU\n"
            " long: East longitude of sub-planet point, deg, positive towards East\n"
            " loc. t.= UT + long: **rough** local solar time at sub-planet point, hh:mm\n"
            " ---------------------------------------------------------------------------------------------------------------------------------------\n"
            " Selection criteria:\n"
            " Maximum corrected magnitudes R* and K* considered: 100.0\n"
            " Maximum geocentrique closest approach considered: {max_ca:.3f} \n"
            " List below has:          {size}  entries\n"
            " Offset applied to ephemeris (mas):     0.0     0.0\n"
            " {radius}                                                Radius (km)\n"
            " {ow_des}          {ow_des}          {name}             Provisional Designation, Number, Name\n"
            " ----------------------------------------------------------------------------------------------------------------------------------------------------\n"
            " d  m year  h:m:s UT    ra___dec___J2000_candidate    ra___dec___J2000_ephemeris     C/A    P/A     vel   Delta   G*   K*   long  loc.t.  e_ra  e_dec\n"
            " ----------------------------------------------------------------------------------------------------------------------------------------------------\n"
              )



from .config import test_attr
from .star import Star
from .ephem import EphemPlanete, EphemJPL, EphemKernel
from .observer import Observer
from .lightcurve import LightCurve
from .prediction import occ_params, PredictionTable
from .extra import ChiSquare
import astropy.units as u
from astropy.coordinates import SkyCoord, SkyOffsetFrame
from astropy.time import Time
import numpy as np
import warnings
import matplotlib.pyplot as plt


def positionv(star,ephem,observer,time):
    """ Calculates the position and velocity of the occultation shadow relative to the observer.
        
    Parameters:
        star (Star): The coordinate of the star in the same frame as the ephemeris.
        It must be a Star object.
        ephem (Ephem): Ephemeris. It must be an Ephemeris object.
        observer (Observer): The Observer object to be added.
        time (Time): Instant to calculate position and velocity
    
    Return:
        f, g (float): The orthographic projection of the shadow relative to the observer
    """
    if type(star) != Star:
        raise ValueError('star must be a Star object')
    if type(ephem) not in [EphemPlanete, EphemJPL, EphemKernel]:
        raise ValueError('ephem must be an Ephemeris object')
    if type(observer) != Observer:
        raise ValueError('observer must be an Observer object')
    time = Time(time)

    coord = star.geocentric(time)
    dt = 0.1*u.s
    
    if type(ephem) == EphemPlanete:
        ephem.fit_d2_ksi_eta(coord, log=False)
    ksio1, etao1 = observer.get_ksi_eta(time=time, star=coord)
    ksie1, etae1 = ephem.get_ksi_eta(time=time, star=coord)
    
    f = ksio1-ksie1
    g = etao1-etae1
    
    ksio2, etao2 = observer.get_ksi_eta(time=time+dt, star=coord)
    ksie2, etae2 = ephem.get_ksi_eta(time=time+dt, star=coord)
    
    nf = ksio2-ksie2
    ng = etao2-etae2
    
    vf = (nf-f)/0.1
    vg = (ng-g)/0.1

    return f, g, vf, vg


def fit_ellipse(*args, **kwargs):
    """ Fits an ellipse to given occultation

    Parameters:
        required params:
        Each occultation is added as the first arguments directly.
        center_f (int,float): The coordinate in f of the ellipse.
        center_g (int,float): The coordinate in g of the ellipse.
        equatorial_radius (int,float): The Equatorial radius of the ellipse.
        oblateness (int,float): The oblateness of the ellipse.
        pos_angle (int,float): The pole position angle of the ellipse.

        Params for the interval of search, if not given, default is set to zero.
        Search between (value - dvalue) and (value + dvalue):
        dcenter_f (int,float)
        dcenter_g (int,float)
        dequatorial_radius (int,float)
        doblateness (int,float)
        dpos_angle (int,float)

        loop: The number of ellipsis to attempt fitting. Default: 10,000,000
        dchi_min: If given, it will only save ellipsis which chi square are
            smaller than chi_min + dchi_min.
        number_chi: if dchi_min is given, the procedure is repeated until
            number_chi is reached. Default: 10,000
        log: If True, it prints information while fitting. Default: False.

    Return:
        chisquare: A ChiSquare object with all parameters.

    Example:
        fit_ellipse(occ1, **kwargs) to fit the ellipse to the chords of occ1 Occultation object
        fit_ellipse(occ1, occ2, **kwargs) to fit the ellipse to the chords
            of occ1 and occ2 Occultation objects together
    """
    params_needed = ['center_f', 'center_g', 'equatorial_radius', 'oblateness', 'pos_angle']
    if not all([param in kwargs for param in params_needed]):
        raise ValueError('Input conditions not satisfied. Please refer to the tutorial.')
    center_f = kwargs['center_f']
    dcenter_f = kwargs.get('dcenter_f', 0.0)
    center_g = kwargs['center_g']
    dcenter_g = kwargs.get('dcenter_g', 0.0)
    equatorial_radius = kwargs['equatorial_radius']
    dequatorial_radius = kwargs.get('dequatorial_radius', 0.0)
    oblateness = kwargs['oblateness']
    doblateness = kwargs.get('doblateness', 0.0)
    pos_angle = kwargs['pos_angle']
    dpos_angle = kwargs.get('dpos_angle', 0.0)
    loop = kwargs.get('loop', 10000000)
    number_chi = kwargs.get('number_chi', 10000)
    log = kwargs.get('log', False)

    values = []
    for occ in args:
        if type(occ) != Occultation:
            raise TypeError('Given argument must be an Occultation object.')
        pos = occ.positions
        for site in pos.keys():
            pos_obs = pos[site]
            for lc in pos_obs.keys():
                pos_lc = pos_obs[lc]
                if type(pos_lc) != _PositionDict:
                    continue
                if pos_lc['status'] == 'positive':
                    if pos_lc['immersion']['on']:
                        f,g = pos_lc['immersion']['value']
                        err = np.array(pos_lc['immersion']['error'])
                        erro = np.linalg.norm(err[0]-err[1])/2.0
                        values.append([f,g,erro])
                    if pos_lc['emersion']['on']:
                        f,g = pos_lc['emersion']['value']
                        err = np.array(pos_lc['emersion']['error'])
                        erro = np.linalg.norm(err[0]-err[1])/2.0
                        values.append([f,g,erro])

    controle_f0 = Time.now()
    f0_chi = np.array([])
    g0_chi = np.array([])
    a_chi  = np.array([])
    obla_chi  = np.array([])
    posang_chi  = np.array([])
    chi2_best = np.array([])

    while (len(f0_chi) < number_chi):
        chi2 = np.zeros(loop)
        f0 = center_f + dcenter_f*(2*np.random.random(loop) - 1)
        g0 = center_g + dcenter_g*(2*np.random.random(loop) - 1)
        a  = equatorial_radius + dequatorial_radius*(2*np.random.random(loop) - 1)
        obla  = oblateness + doblateness*(2*np.random.random(loop) - 1)
        obla[obla<0],obla[obla>1] = 0, 1
        phi_deg  = pos_angle + dpos_angle*(2*np.random.random(loop) - 1)
        controle_f1 = Time.now()

        for fi, gi, si in values:
            b = a - a*obla
            phi = phi_deg*(np.pi/180.0)
            dfi = fi-f0
            dgi = gi-g0
            r = np.sqrt(dfi**2 + dgi**2)
            theta = np.arctan2(dgi,dfi)
            ang = theta+phi
            r_model = (a*b)/np.sqrt((a*np.sin(ang))**2 + (b*np.cos(ang))**2)
            f_model = f0 + r_model*np.cos(theta)
            g_model = g0 + r_model*np.sin(theta)
            chi2 += ((fi - f_model)**2 + (gi - g_model)**2)/(si**2)

        controle_f2 = Time.now()
        if 'dchi_min' in kwargs:
            region = np.where(chi2 < chi2.min() + kwargs['dchi_min'])[0]
        else:
            region = np.arange(len(chi2))
        chi2_best = np.append(chi2_best,chi2[region])
        if log:
            print('Elapsed time: {:.3f} seconds.'.format((controle_f2 - controle_f1).sec))
            print(len(chi2[region]),len(chi2_best)) #TEST
        f0_chi = np.append(f0_chi,f0[region])
        g0_chi = np.append(g0_chi,g0[region])
        a_chi  = np.append(a_chi,a[region])
        obla_chi  = np.append(obla_chi,obla[region])
        posang_chi  = np.append(posang_chi,phi_deg[region])

    chisquare = ChiSquare(chi2_best, len(values), center_f=f0_chi, center_g=g0_chi, equatorial_radius=a_chi,
                          oblateness=obla_chi, position_angle=posang_chi)
    controle_f4 = Time.now()
    if log:
        print('Total elapsed time: {:.3f} seconds.'.format((controle_f4 - controle_f0).sec))

    onesigma = chisquare.get_nsigma(sigma=1)
    for occ in args:
        if type(occ) == Occultation:
            occ.fitted_params = {i:onesigma[i] for i in ['equatorial_radius', 'center_f', 'center_g', 'oblateness', 'position_angle']}
    a = occ.fitted_params['equatorial_radius'][0]
    f0 = occ.fitted_params['center_f'][0]
    g0 = occ.fitted_params['center_g'][0]
    obla = occ.fitted_params['oblateness'][0]
    phi_deg = occ.fitted_params['position_angle'][0]
    radial_dispersion = np.array([])
    error_bar = np.array([])
    for fi, gi, si in values:
        b = a - a*obla
        phi = phi_deg*(np.pi/180.0)
        dfi = fi-f0
        dgi = gi-g0
        r = np.sqrt(dfi**2 + dgi**2)
        theta = np.arctan2(dgi,dfi)
        ang = theta+phi
        r_model = (a*b)/np.sqrt((a*np.sin(ang))**2 + (b*np.cos(ang))**2)
        f_model = f0 + r_model*np.cos(theta)
        g_model = g0 + r_model*np.sin(theta)
        radial_dispersion = np.append(radial_dispersion,r - r_model)
        error_bar = np.append(error_bar,si)
    occ.chi2_params = {'radial_dispersion': [radial_dispersion.mean(),radial_dispersion.std(ddof=1)]}
    occ.chi2_params['mean_error'] = [error_bar.mean(),error_bar.std()]
    occ.chi2_params['chi2_min'] = chisquare.get_nsigma()['chi2_min']
    occ.chi2_params['nparam'] = chisquare.nparam
    occ.chi2_params['npts'] = chisquare.npts
    return chisquare


class _PositionDict(dict):
    """ This is a modified Dictionary object to allow for switch on/off of data points,
    it also avoids for user to change data.
    """
    def __setitem__(self, key, value):
        """Redefine how to set a value to a key in the dictionary.
        it only sets a value if the key starts with '_occ_'.
        Otherwise, it only allows for the user to provide 'on' or 'off'
        which is passed only to change the 'on' keyword.
        """
        status = {'on': True, 'off': False}
        n = 0
        if key.startswith('_occ_'):
            super().__setitem__(key[5:], value)
            n = 1
        elif key in self.keys():
            n = 1
            if value not in status.keys():
                raise ValueError("Value must be 'on' or 'off' only.")
            if type(self[key]) == _PositionDict:
                for k in self[key].keys():
                    self[key][k] = value
            elif key=='on':
                super().__setitem__('on', status[value])
        else:
            if value not in status.keys():
                raise ValueError("Value must be 'on' or 'off' only.")
            for key1 in self.keys():
                if type(self[key1]) == _PositionDict:
                    if key in self[key1].keys():
                        n = 1
                        self[key1][key] = value
        if n==0:
            raise KeyError('Key "{}" does not exist'.format(key))

    def __str__(self):
        out = '\n' + '\n'.join(['{}: {}'.format(key, self[key]) for key in self.keys()])
        return out.replace('\n', '\n  ')


### Object for occultation
class Occultation():
    '''
    Docstring
    Do the reduction of the occultation
    '''
    def __init__(self, star, ephem, time):
        """ Instantiate Occultation object.

        Parameters:
            star (Star):The coordinate of the star in the same frame as the ephemeris.
            It must be a Star object.
            ephem (Ephem):Ephemeris. It must be an Ephemeris object.
            time (str, Time): The approximate time of the occultation (~10min)
                to calculate occultation params.
        """
        if type(star) != Star:
            raise ValueError('star must be a Star object')
        if type(ephem) not in [EphemPlanete, EphemKernel, EphemJPL]:
            raise ValueError('ephem must be a Ephemeris object')
        self.star = star
        self.ephem = ephem
        
        tca, ca, pa, vel, dist = occ_params(star,ephem, time)
        self.ca = ca   # Closest Approach distance
        self.pa = pa   # Position Angle at CA
        self.vel = vel  # Shadow velocity at CA
        self.dist = dist  # object distance at CA
        self.tca = tca   # Instant of CA
        self.star_diam = self.star.apparent_diameter(self.dist, log=False)

        meta = {'name': self.ephem.name, 'radius': self.ephem.radius.to(u.km).value,
            'error_ra': self.ephem.error_ra.to(u.mas).value, 'error_dec': self.ephem.error_dec.to(u.mas).value}
        self.predict = PredictionTable(time=[tca], coord_star=[self.star.geocentric(tca)],
            coord_obj=[self.ephem.get_position(tca)], ca=[ca.value], pa=[pa.value], vel=[vel.value],
            dist=[dist.value], mag=[self.star.mag['G']], source=[self.star.code], meta=meta)
        
        self.__observations = []
        self._position = _PositionDict()
    
    def add_observation(self, obs, lightcurve):
        """ Add observations to the Occultation object.

        Parameters:
            obs (Observer):The Observer object to be added.
            status (string): it can be "positive", "negative", "visual" or "undefined"
        """
        if type(obs) != Observer:
            raise ValueError('obs must be an Observer object')
        if type(lightcurve) != LightCurve:
            raise ValueError('lightcurve must be a LightCurve object')
        for o,l in self.__observations:
            if l.name == lightcurve.name:
                raise ValueError('{} LightCurve already associated to {} Observer'.format(lightcurve.name, o.name))
        self.__observations.append((obs,lightcurve))
        lightcurve.set_vel(np.absolute(self.vel))
        lightcurve.set_dist(float(self.dist.AU))
        lightcurve.set_star_diam(float(self.star_diam.km))
        try:
            lightcurve.calc_magnitude_drop(mag_star=self.star.mag['G'],mag_obj=self.ephem.apparent_magnitude(self.tca))
        except:
            lightcurve.bottom_flux = 0.0
            warnings.warn('Magnitude drop was not calculated. Using bottom flux as 0.0 instead.')


    def remove_observation(self, key, key_lc=None):
        """ Removes observation from the Occultation object.

        Parameters:
            key: The name given to Observer or LightCurve to remove from the list.
            keylc: In the case where repeated names are present for different observations,
                keylc must be given for the name of the LightCurve
                and key will be used for the name of the Observer.
        """
        rm_list = np.array([])
        obs = []
        lcs = []
        same_key = False
        if key_lc is None:
            same_key = True
            key_lc = key
        ko = []
        kl = []
        for i, val in enumerate(self.__observations):
            if val[0].name == key:
                ko.append(i)
            if val[1].name == key_lc:
                kl.append(i)
        ko = np.array(ko)
        kl = np.array(kl)
        if not same_key:
            k  = ko[np.where(ko == kl)[0]]
            rm_list = np.hstack((rm_list, k))
        else:
            rm_list = np.hstack((rm_list, np.array(kl)))
            if len(kl) > 0 and len(ko) > 0 and kl[0] not in ko:
                raise ValueError("Observation could not univocally be identified, please give parameters for Observer and LightCurve")
            rm_list = np.hstack((rm_list, np.array(ko)))
        rm_list = np.unique(rm_list)
        if len(rm_list) == 0:
            raise ValueError('No observer "{}" and/or lightcurve "{}" was found'.format(key,key_lc))
        list = np.arange(len(self.__observations)).tolist()
        for i in rm_list:
            list.remove(i)
        self.__observations = [self.__observations[item] for item in list]

    def observations(self):
        """ Print all the observations added to the Occultation object
        Pair (Observer, LightCurve)
        """
        for o,l in self.__observations:
            print('Observer= {}, LC: {}'.format(o.name, l.name))

    def fit_ellipse(self, **kwargs):
        """ Fits an ellipse to the chords of this occultation

        Parameters:
            required params:
            Each occultation is added as the first arguments directly.
            center_f (int,float): The coordinate in f of the ellipse.
            center_g (int,float): The coordinate in g of the ellipse.
            equatorial_radius (int,float): The Equatorial radius of the ellipse.
            oblateness (int,float): The oblateness of the ellipse.
            pos_angle (int,float): The pole position angle of the ellipse.

            Params for the interval of search, if not given, default is set to zero.
            Search between (value - dvalue) and (value + dvalue):
            dcenter_f (int,float)
            dcenter_g (int,float)
            dequatorial_radius (int,float)
            doblateness (int,float)
            dpos_angle (int,float)

            loop: The number of ellipsis to attempt fitting. Default: 10,000,000
            dchi_min: If given, it will only save ellipsis which chi square are
                smaller than chi_min + dchi_min.
            number_chi: if dchi_min is given, the procedure is repeated until
                number_chi is reached. Default: 10,000
            log: If True, it prints information while fitting. Default: False.

        Return:
            chisquare: A ChiSquare object with all parameters.
        """
        chisquare = fit_ellipse(self, **kwargs)
        return chisquare

    @property
    def positions(self):
        """ Calculates the positions and velocities for all chords.
        Saves it into an _PositionDict object.
        """
        position = self._position
        if len(self.__observations) == 0:
            raise ValueError('There is no observation defined for this occultation')

        pair = []
        for o,l in self.__observations:
            pair.append((o.name, l.name))

            coord = [o.lon,o.lat,o.height]
            if o.name not in position.keys():
                position['_occ_'+o.name] = _PositionDict(lon=o.lon, lat=o.lat, height=o.height)
                position[o.name]['_occ_lon'] = o.lon
                position[o.name]['_occ_lat'] = o.lat
                position[o.name]['_occ_height'] = o.height
            pos_obs = position[o.name]
            coord2 = [pos_obs['lon'],pos_obs['lat'],pos_obs['height']]
            if o.lon != pos_obs['lon']:
                position[o.name]['_occ_lon'] = o.lon
            if o.lat != pos_obs['lat']:
                position[o.name]['_occ_lat'] = o.lat
            if o.height != pos_obs['height']:
                position[o.name]['_occ_height'] = o.height
            samecoord = (coord == coord2)

            if l.name not in pos_obs.keys():
                pos_obs['_occ_'+l.name] = _PositionDict()
            pos_lc = pos_obs[l.name]

            pos_lc['_occ_status'] = 'negative'
            if hasattr(l, 'immersion') or hasattr(l, 'emersion'):
                pos_lc['_occ_status'] = 'positive'

            if hasattr(l, 'immersion'):
                if 'immersion' not in pos_lc.keys():
                    pos_lc['_occ_immersion'] = _PositionDict(on=True)
                obs_im = pos_lc['immersion']
                do_err = False
                if samecoord and 'time' in obs_im.keys() and obs_im['time'] == l.immersion:
                    pass
                else:
                    do_err = True
                    f1,g1,vf1,vg1 = positionv(self.star,self.ephem,o,l.immersion)
                    obs_im['_occ_time'] =l.immersion
                    obs_im['_occ_value'] = (round(f1,3),round(g1,3))
                    obs_im['_occ_vel'] = (round(vf1,3),round(vg1,3))
                if not do_err and 'time_err' in obs_im.keys() and obs_im['time_err'] == l.immersion_err:
                    pass
                else:
                    fe1,ge1 = positionv(self.star,self.ephem,o,l.immersion-l.immersion_err*u.s)[0:2]
                    fe2,ge2 = positionv(self.star,self.ephem,o,l.immersion+l.immersion_err*u.s)[0:2]
                    obs_im['_occ_time_err'] = l.immersion_err
                    obs_im['_occ_error'] = ((round(fe1,3),round(ge1,3)),(round(fe2,3),round(ge2,3)))

            if hasattr(l, 'emersion'):
                if 'emersion' not in pos_lc.keys():
                    pos_lc['_occ_emersion'] = _PositionDict(on=True)
                obs_em = pos_lc['emersion']
                do_err = False
                if samecoord and 'time' in obs_em.keys() and obs_em['time'] == l.emersion:
                    pass
                else:
                    do_err = True
                    f1,g1,vf1,vg1 = positionv(self.star,self.ephem,o,l.emersion)
                    obs_em['_occ_time'] =l.emersion
                    obs_em['_occ_value'] = (round(f1,3),round(g1,3))
                    obs_em['_occ_vel'] = (round(vf1,3),round(vg1,3))
                if not do_err and 'time_err' in obs_em.keys() and obs_em['time_err'] == l.emersion_err:
                    pass
                else:
                    fe1,ge1 = positionv(self.star,self.ephem,o,l.emersion-l.emersion_err*u.s)[0:2]
                    fe2,ge2 = positionv(self.star,self.ephem,o,l.emersion+l.emersion_err*u.s)[0:2]
                    obs_em['_occ_time_err'] = l.emersion_err
                    obs_em['_occ_error'] = ((round(fe1,3),round(ge1,3)),(round(fe2,3),round(ge2,3)))

            if pos_lc['status'] == 'negative':
                if 'start_obs' not in pos_lc.keys():
                    pos_lc['_occ_start_obs'] = _PositionDict(on=True)
                obs_start = pos_lc['start_obs']
                if samecoord and 'time' in obs_start.keys() and obs_start['time'] == l.initial_time:
                    pass
                else:
                    f,g, vf,vg = positionv(self.star,self.ephem,o,l.initial_time)
                    obs_start['_occ_time'] = l.initial_time
                    obs_start['_occ_value'] = (round(f,3),round(g,3))
                    obs_start['_occ_vel'] = (round(vf,3),round(vg,3))
                if 'end_obs' not in pos_lc.keys():
                    pos_lc['_occ_end_obs'] = _PositionDict(on=True)
                obs_end = pos_lc['end_obs']
                if samecoord and 'time' in obs_end.keys() and obs_end['time'] == l.end_time:
                    pass
                else:
                    f,g,vf,vg = positionv(self.star,self.ephem,o,l.end_time)
                    obs_end['_occ_time'] = l.end_time
                    obs_end['_occ_value'] = (round(f,3),round(g,3))
                    obs_end['_occ_vel'] = (round(vf,3),round(vg,3))

        for key in list(position):
            n = 0
            for key_lc in list(position[key]):
                if type(position[key][key_lc]) != _PositionDict:
                    continue
                if (key,key_lc) not in pair:
                    del position[key][key_lc]
                else:
                    n+=1
            if n == 0:
                del position[key]

        return self._position

    @positions.setter
    def positions(self, value):
        """ If the users tries to set a value to position, it must be 'on' or 'off',
        and it will be assigned to all chords.
        """
        if not hasattr(self, '_position'):
            pos = self.positions
        if value not in ['on','off']:
            raise ValueError("Value must be 'on' or 'off' only.")
        for key in self._position.keys():
            self._position[key] = value

    def check_velocities(self):
        """ Print the current velocity used by the LightCurves and the Radial velocity.
        """
        if hasattr(self, 'fitted_params'):
            center = np.array([self.fitted_params['center_f'][0], self.fitted_params['center_g'][0]])
        else:
            center = np.array([0,0])
        positions = self.positions
        for o,l in self.__observations:
            vals = positions[o.name][l.name]
            if all([i not in vals.keys() for i in ['immersion', 'emersion']]):
                continue
            print('{} - Velocity used: {:.3f}'.format(l.name, l.vel))
            if 'immersion' in vals.keys():
                im = vals['immersion']
                delta = np.array(im['value']) - center
                print('    Immersion Radial Velocity: {:.3f}'.
                      format(np.abs(np.dot(np.array(im['vel']), delta)/np.linalg.norm(delta))))
            if 'emersion' in vals.keys():
                em = vals['emersion']
                delta = np.array(em['value']) - center
                print('    Emersion Radial Velocity: {:.3f}'.
                      format(np.abs(np.dot(np.array(em['vel']), delta)/np.linalg.norm(delta))))

    def new_astrometric_position(self, time=None, offset=None, error=None, log=True):
        """ Calculates the new astrometric position for the object given fitted params

        Parameters:
            time: Time to which calculate the position. If not given, it uses the instant at C/A.
            offset (list): Offset to apply to the position. If not given, it uses the params fitted from ellipse.
                Offsets must be a list of 3 values being [X, Y, 'unit'], where 'unit' must be an angular or distance unit.
                Angular units must be in dacosdec, ddec: Ex: [30.6, 20, 'mas'], or [-15, 2, 'arcsec']
                Distance units must be in X and Y: Ex: [100, -200, 'km'], [0.001, 0.002, 'AU']
            error (list): Error bar of the given offset. If not given, it uses the 1-sigma fitted from ellipse.
                Error must be a list of 3 values being [dX, dY, 'unit'], similar to offset.
                It does not need to be in the same unit as offset.
            log (bool): If true, it prints text, else it returns text.
        """
        if time is not None:
            time = Time(time)
        else:
            time = self.tca

        if offset is not None:
            off_ra = offset[0]*u.Unit(offset[2])
            off_dec = offset[1]*u.Unit(offset[2])
            try:
                teste = off_ra.to(u.km)
                dist = True
            except:
                try:
                    teste = off_ra.to(u.arcsec)
                    dist = False
                except:
                    raise ValueError('Offset unit must be a distance or angular value.')
        elif hasattr(self, 'fitted_params'):
            off_ra = self.fitted_params['center_f'][0]*u.km
            off_dec = self.fitted_params['center_g'][0]*u.km
            dist = True
        else:
            warnings.warn('No offset given or found. Using 0.0 instead.')
            off_ra = 0.0*u.mas
            off_dec = 0.0*u.mas
            dist = False

        if error is not None:
            e_off_ra = error[0]*u.Unit(error[2])
            e_off_dec = error[1]*u.Unit(error[2])
            try:
                teste = e_off_ra.to(u.km)
                e_dist = True
            except:
                try:
                    teste = e_off_ra.to(u.arcsec)
                    e_dist = False
                except:
                    raise ValueError('Error unit must be a distance or angular value.')
        elif hasattr(self, 'fitted_params'):
            e_off_ra = self.fitted_params['center_f'][1]*u.km
            e_off_dec = self.fitted_params['center_g'][1]*u.km
            e_dist = True
        else:
            warnings.warn('No error given or found. Using 0.0 instead.')
            e_off_ra = 0.0*u.mas
            e_off_dec = 0.0*u.mas
            e_dist = False

        coord_geo = self.ephem.get_position(time)
        distance = coord_geo.distance.to(u.km)
        coord_frame = SkyOffsetFrame(origin=coord_geo)
        if dist:
            off_ra = np.arctan2(off_ra, distance)
            off_dec = np.arctan2(off_dec, distance)
        if e_dist:
            e_off_ra = np.arctan2(e_off_ra, distance)
            e_off_dec = np.arctan2(e_off_dec, distance)
        new_pos = SkyCoord(lon=off_ra, lat=off_dec, frame=coord_frame)
        new_pos = new_pos.icrs

        error_star = self.star.error_at(self.tca)
        error_ra = error_star[0] + e_off_ra
        error_dec = error_star[1] + e_off_dec

        out = 'Ephemeris offset (km): X = {:.1f} +/- {:.1f}; Y = {:.1f} +/- {:.1f}\n'.format(
              distance*np.sin(off_ra.to(u.mas)).value, distance*np.sin(e_off_ra.to(u.mas)).value,
              distance*np.sin(off_dec.to(u.mas)).value, distance*np.sin(e_off_dec.to(u.mas)).value)
        out += 'Ephemeris offset (mas): da_cos_dec = {:.3f} +/- {:.3f}; d_dec = {:.3f} +/- {:.3f}\n'.format(
              off_ra.to(u.mas).value, e_off_ra.to(u.mas).value, off_dec.to(u.mas).value, e_off_dec.to(u.mas).value)
        out += '\nAstrometric object position at time {}\n'.format(time.iso)
        out += 'RA = {} +/- {:.3f} mas; DEC = {} +/- {:.3f} mas'.format(
               new_pos.ra.to_string(u.hourangle, precision=5, sep=' '), error_ra.to(u.mas).value,
               new_pos.dec.to_string(u.deg, precision=4, sep=' '), error_dec.to(u.mas).value)

        if log:
            print(out)
        else:
            return out

    def plot_chords(self, all_chords=True, positive_color='blue', negative_color='green', error_color='red', ax=None):
        """Plots the chords of the occultation

        Parameters:
            all_chords (bool): if True, it plots all the chords,
                if False, it sees what was deactivated in self.positions and ignores them
            positive_color: color for the positive chords. Default: blue
            negative_color: color for the negative chords. Default: green
            error_color: color for the error bars of the chords. Default: red
        """
        ax = ax or plt.gca()
        positions = self.positions
        for site in positions.keys():
            pos_obs = positions[site]
            for lc in pos_obs.keys():
                pos_lc = pos_obs[lc]
                if type(pos_lc) != _PositionDict:
                    continue
                if pos_lc['status'] == 'negative':
                    arr = np.array([pos_lc['start_obs']['value'], pos_lc['end_obs']['value']])
                    ax.plot(*arr.T, '--', color=negative_color, linewidth=0.7)
                else:
                    n = 0
                    if pos_lc['immersion']['on'] or all_chords:
                        arr = np.array([pos_lc['immersion']['error']])
                        ax.plot(*arr.T, color=error_color, linewidth=1.5)
                        n+=1
                    if pos_lc['emersion']['on'] or all_chords:
                        arr = np.array([pos_lc['emersion']['error']])
                        ax.plot(*arr.T, color=error_color, linewidth=1.5)
                        n+=1
                    if n == 2:
                        arr = np.array([pos_lc['immersion']['value'], pos_lc['emersion']['value']])
                        ax.plot(*arr.T, color=positive_color, linewidth=0.7)
        ax.axis('equal')

    def get_map_sites(self):
        """Return Dictionary with sites in the format required by plot_occ_map function

        Returns:
            sites (dict): Dictionary with the sites in the format required by plot_occ_map function
        """
        sites = {}
        color = {'positive': 'blue', 'negative': 'red'}
        positions = self.positions
        for o, l in self.__observations:
            sites[o.name] = [o.lon.deg, o.lat.deg, 10, 10, color[positions[o.name][l.name]['status']]]
        return sites

    def plot_occ_map(self, **kwargs):
        """Plot occultation map

        Parameters:
            radius: The radius of the shadow. If not given it uses the equatorial radius
                from an ellipse fit, else it uses the radius obtained from ephem upon
                instantiating.

            nameimg (str): Change the name of the imaged saved.
            resolution (int): Cartopy feature resolution. "1" means a resolution of "10m",
                "2" a resolution of "50m" and "3" a resolution of "100m". Default = 2
            states (bool): True to plot the state division of the countries. The states of
                some countries will only be shown depending on the resolution.
            zoom (int, float): Zooms in or out of the map.
            centermap_geo: Center the map given coordinates in longitude and latitude.
                It must be a list with two numbers. Default=None.
            centermap_delta: Displace the center of the map given displacement
                in X and Y, in km. It must be a list with two numbers. Default=None.
            centerproj: Rotates the Earth to show occultation with the center
                projected at a given longitude and latitude.
            labels: Plots text above and below the map with the occultation parameters.
                Default=True.
            meridians: Plots lines representing the meridians for given interval. Default=30 deg
            parallels: Plots lines representing the parallels for given interval. Default=30 deg
            sites: Plots site positions in map. It must be a python dictionary where the key is
                the name of the site, and the value is a list with longitude, latitude, delta_x,
                delta\_y and color. delta_x and delta_y are displacement, in km, from the point
                of the site in the map and the name. color is the color of the point.
                If not given, it calculates from observations added to Occultation
            countries: Plots the names of countries. It must be a python dictionary where the key
                is the name of the country and the value is a list with longitude and latitude
                of the lower left part of the text.
            offset: applies an offset to the ephemeris, calculating new CA and instant of CA.
                It is a pair of delta_RA*cosDEC and delta_DEC.
                If not given it uses the center from ellipse fitted.
            mapstyle: Define the color style of the map. 1 is the default black and white scale.
                2 is a colored map.
            error: Ephemeris error in mas. It plots a dashed line representing radius + error.
            lncolor: Changes the color of the lines of the error bar.
            ring: It plots a dashed line representing the location of a ring.
                It is given in km, from the center.
            rncolor: Changes the color of ring lines.
            atm: It plots a dashed line representing the location of an atmosphere.
                It is given in km, from the center.
            rncolor: Changes the color of atm lines.
            heights: It plots a circular dashed line showing the locations where the observer
                would observe the occultation at a given height above the horizons.
                This must be a list.
            hcolor: Changes the color of the height lines.
            mapsize: The size of figure, in cm. It must be a list with two values.
                Default = [46.0, 38.0].
            cpoints: Interval for the small points marking the center of shadow,
                in seconds. Default=60.
            ptcolor: Change the color of the center points.
            alpha: The transparency of the night shade, where 0.0 is full transparency
                and 1.0 is full black. Default = 0.2.
            fmt: The format to save the image. It is parsed directly by matplotlib.pyplot.
                Default = 'png'
            dpi: "Dots per inch". It defines the quality of the image. Default = 100.
            lncolor: Changes the color of the line that represents the limits of the shadow over Earth.
            outcolor: Changes the color of the lines that represents the limits of the shadow outside Earth
            nscale: Arbitrary scale for the size for the name of the site.
            cscale Arbitrary scale for the name of the country.
            sscale Arbitrary scale for the size of point of the site.
            pscale: Arbitrary scale for the size of the points that represent the center of the shadow
            arrow (bool): If true, it plots the arrow with the occultation direction.

            Comment: Only one of centermap_geo and centermap_delta can be given
        """
        if 'radius' not in kwargs and hasattr(self, 'fitted_params'):
            kwargs['radius'] = self.fitted_params['equatorial_radius'][0]
        kwargs['sites'] = kwargs.get('sites', self.get_map_sites())
        if 'offset' not in kwargs and hasattr(self, 'fitted_params'):
            off_ra = self.fitted_params['center_f'][0]*u.km
            off_dec = self.fitted_params['center_g'][0]*u.km
            off_ra = np.arctan2(off_ra, self.dist)
            off_dec = np.arctan2(off_dec, self.dist)
            kwargs['offset'] = [off_ra, off_dec]
        self.predict.plot_occ_map(**kwargs)

    def to_log(self,namefile=None):
        """ Save the occultation log to a file

        Parameters:
            namefile (str): Filename to save the log
        """
        if (namefile == None):
            namefile = 'occ_{}_{}.log'.format(self.ephem.name, self.tca.isot[:16])
        f = open(namefile, 'w')
        f.write(self.__str__())
        f.close()

    def to_file(self):
        """ Save the occultation data to a file

        Three files are saved containing the positions and velocities
        for the observations. They are for the positive, negative and
        error bars positions.

        The format of the files are: positions in f and g,
        velocities in f and g, the Julian Date of the observation,
        and light curve name of the corresponding position.
        """
        positions = self.positions
        pos = []
        neg =[]
        err = []
        for o,l in self.__observations:
            status = positions[o.name][l.name]['status']
            l_name = l.name.replace(' ', '_')
            if status == 'positive':
                im = positions[o.name][l.name]['immersion']
                f,g = im['value']
                err1, err2 = im['error']
                f1,g1 = err1
                f2,g2 = err2
                vf,vg = im['vel']
                pos.append([f,g,vf,vg, im['time'].jd, l_name+'_immersion'])
                err.append([f1,g1,vf,vg,(im['time']-im['time_err']*u.s).jd,l_name+'_immersion_err-'])
                err.append([f2,g2,vf,vg,(im['time']+im['time_err']*u.s).jd,l_name+'_immersion_err+'])

                em = positions[o.name][l.name]['emersion']
                f,g = em['value']
                err1, err2 = em['error']
                f1,g1 = err1
                f2,g2 = err2
                vf,vg = em['vel']
                pos.append([f,g,vf,vg, em['time'].jd, l_name+'_emersion'])
                err.append([f1,g1,vf,vg,(em['time']-em['time_err']*u.s).jd,l_name+'_emersion_err-'])
                err.append([f2,g2,vf,vg,(em['time']+em['time_err']*u.s).jd,l_name+'_emersion_err+'])
            if status == 'negative':
                ini = positions[o.name][l.name]['start_obs']
                f,g = ini['value']
                vf,vg = ini['vel']
                neg.append([f,g,vf,vg,ini['time'].jd,l_name+'_start'])

                end = positions[o.name][l.name]['end_obs']
                f,g = end['value']
                vf,vg = end['vel']
                neg.append([f,g,vf,vg,end['time'].jd,l_name+'_end'])
        if len(pos) > 0:
            f = open('occ_{}_pos.txt'.format(self.ephem.name), 'w')
            for line in pos:
                f.write('{:9.3f} {:9.3f} {:-6.2f} {:-6.2f} {:16.8f} {}\n'.format(*line))
            f.close()
            f = open('occ_{}_err.txt'.format(self.ephem.name), 'w')
            for line in err:
                f.write('{:9.3f} {:9.3f} {:-6.2f} {:-6.2f} {:16.8f} {}\n'.format(*line))
            f.close()
        if len(neg) > 0:
            f = open('occ_{}_neg.txt'.format(self.ephem.name), 'w')
            for line in neg:
                f.write('{:9.3f} {:9.3f} {:-6.2f} {:-6.2f} {:16.8f} {}\n'.format(*line))
            f.close()

    def __str__(self):
        """String representation of the Star class
        """
        out = 'Stellar occultation of star Gaia-DR2 {} by {}.\n\n'.format(self.star.code, self.ephem.name)
        out += 'Geocentric Closest Approach: {:.3f}\n'.format(self.ca)
        out += 'Instant of CA: {}\n'.format(self.tca.iso)
        out += 'Position Angle: {:.2f}\n'.format(self.pa)
        out += 'Geocentric shadow velocity: {:.2f}\n\n\n'.format(self.vel)

        count = {'positive': 0, 'negative': 0, 'visual': 0}
        string = {'positive': '', 'negative': '', 'visual': ''}
        if len(self.__observations) > 0:
            pos = self.positions
            for o,l in self.__observations:
                status = pos[o.name][l.name]['status']
                if count[status] > 0:
                    string[status] += '-'*79 + '\n'
                string[status] += o.__str__() + '\n\n'
                string[status] += l.__str__() + ''
                count[status] += 1

        if np.sum([count[i] for i in count.keys()]) == 0:
            out += 'No observations reported'
        else:
            out += '\n'.join(['{} {} observations'.format(count[k],k) for k in string.keys() if count[k] > 0])

        out += '\n\n'

        out += '#'*79 + '\n{:^79s}\n'.format('STAR') + '#'*79 + '\n'
        out += self.star.__str__() + '\n\n'

        coord = self.star.geocentric(self.tca)
        error_star = self.star.error_at(self.tca)
        out += 'GCRS star coordinate at occultation Epoch ({}):\n'.format(self.tca.iso)
        out += 'RA={} +/- {:.4f}, DEC={} +/- {:.4f}\n\n'.format(
            coord.ra.to_string(u.hourangle, sep='hms', precision=5), error_star[0],
            coord.dec.to_string(u.deg, sep='dms', precision=4), error_star[1])
        
        out += '#'*79 + '\n{:^79s}\n'.format('EPHEMERIS') + '#'*79 +'\n'
        out += self.ephem.__str__() + '\n'

        for status in string.keys():
            if count[status] == 0:
                continue
            out += ('#'*79 + '\n{:^79s}\n'.format(status.upper() + ' OBSERVATIONS') + '#'*79)
            out += '\n\n'
            out += string[status]

        if hasattr(self, 'fitted_params'):
            out += '#'*79 + '\n{:^79s}\n'.format('RESULTS') + '#'*79 + '\n\n'
            out += 'Fitted Ellipse:\n'
            out += '\n'.join(['{}: {:.3f} +/- {:.3f}'.format(k, *self.fitted_params[k])
                              for k in self.fitted_params.keys()]) + '\n'
            out += '\nMinimum chi-square: {:.3f}\n'.format(self.chi2_params['chi2_min'])
            out += 'Number of fitted points: {}\n'.format(self.chi2_params['npts'])
            out += 'Number of fitted parameters: {}\n'.format(self.chi2_params['nparam'])
            out += 'Minimum chi-square per degree of freedom: {:.3f}\n'.format(self.chi2_params['chi2_min']/ (self.chi2_params['npts'] - self.chi2_params['nparam']))
            out += 'Radial dispersion: {:.3f} +/- {:.3f} km\n'.format(self.chi2_params['radial_dispersion'][0], self.chi2_params['radial_dispersion'][1])
            out += 'Mean error: {:.3f} +/- {:.3f} km\n'.format(self.chi2_params['mean_error'][0], self.chi2_params['mean_error'][1])
            
            out += '\n' + self.new_astrometric_position(log=False)

        return out
